from pathlib import Path
import unittest

import numpy as np


from model_server.base.accessors import generate_file_accessor
from model_server.base.api import app
import model_server.conf.testing as conf
from model_server.base.pipelines.roiset_obmap import RoiSetObjectMapParams, roiset_object_map_pipeline
import model_server.extensions.ilastik.models as ilm
from model_server.extensions.ilastik.router import router

app.include_router(router)

data = conf.meta['image_files']
output_path = conf.meta['output_path']
test_params = conf.meta['roiset']
classifiers = conf.meta['ilastik_classifiers']


class BaseTestRoiSetMonoProducts(object):

    @property
    def fpi(self):
        return data['multichannel_zstack_raw']['path'].__str__()

    @property
    def stack(self):
        return generate_file_accessor(self.fpi)

    @property
    def stack_ch_pa(self):
        return self.stack.get_mono(test_params['patches_channel'])

    @property
    def seg_mask(self):
        return generate_file_accessor(data['multichannel_zstack_mask2d']['path'])

    def _get_export_params(self):
        return {
            'patches_3d': None,
            'annotated_patches_2d': {
                'draw_bounding_box': True,
                'rgb_overlay_channels': [3, None, None],
                'rgb_overlay_weights': [0.2, 1.0, 1.0],
                'pad_to': 512,
            },
            'patches_2d': {
                'draw_bounding_box': False,
                'draw_mask': False,
            },
            'annotated_zstacks': None,
            'object_classes': True,
        }

    def _get_roi_params(self):
        return {
            'mask_type': 'boxes',
            'filters': {
                'area': {'min': 1e0, 'max': 1e8}
            },
            'expand_box_by': [128, 2],
            'deproject_channel': 0,
        }

    def _get_models(self): # tests can either use model objects directly, or load in API via project file string
        fp_px = classifiers['px']['path'].__str__()
        fp_ob = classifiers['seg_to_obj']['path'].__str__()
        return {
            'pixel_classifier_segmentation': {
                'name': 'ilastik_px_mod',
                'project_file': fp_px,
                'model': ilm.IlastikPixelClassifierModel(
                    ilm.IlastikPixelClassifierParams(
                        project_file=fp_px,
                    )
                )
            },
            'object_classifier': {
                'name': 'ilastik_ob_mod',
                'project_file': fp_ob,
                'model': ilm.IlastikObjectClassifierFromSegmentationModel(
                    ilm.IlastikParams(
                        project_file=fp_ob
                    )
                )
            },
        }


class TestRoiSetWorkflow(BaseTestRoiSetMonoProducts, unittest.TestCase):

    def _pipeline_params(self):
        return {
            'api': False,
            'accessor_id': 'acc_id',
            'pixel_classifier_segmentation_model_id': 'px_id',
            'object_classifier_model_id': 'ob_id',
            'segmentation': {
                'channel': test_params['segmentation_channel'],
            },
            'patches_channel': test_params['patches_channel'],
            'roi_params': self._get_roi_params(),
            'export_params': self._get_export_params(),
        }

    def test_object_map_workflow(self):
        acc_in = generate_file_accessor(self.fpi)
        params = RoiSetObjectMapParams(
            **self._pipeline_params(),
        )
        trace, rois = roiset_object_map_pipeline(
            {'accessor': acc_in},
            {f'{k}_model': v['model'] for k, v in self._get_models().items()},
            **params.dict()
        )
        self.assertEqual(trace.pop('annotated_patches_2d').count, 13)
        self.assertEqual(trace.pop('patches_2d').count, 13)
        trace.write_interm(Path(output_path) / 'trace', 'roiset_worfklow_trace', skip_first=False, skip_last=False)
        self.assertTrue('ob_id' in trace.keys())
        self.assertEqual(len(trace['labeled'].unique()[0]), 14)
        self.assertEqual(rois.count, 13)
        self.assertEqual(len(trace['ob_id'].unique()[0]), 2)


class TestRoiSetWorkflowOverApi(conf.TestServerBaseClass, BaseTestRoiSetMonoProducts):

    app_name = 'tests.test_ilastik.test_roiset_workflow:app'
    input_data = data['multichannel_zstack_raw']


    def setUp(self) -> None:
        self.where_out = output_path / 'roiset'
        self.where_out.mkdir(parents=True, exist_ok=True)
        return conf.TestServerBaseClass.setUp(self)

    def test_trivial_api_response(self):
        self.assertGetSuccess('')

    def test_load_input_accessor(self):
        fname = self.copy_input_file_to_server()
        return self.assertPutSuccess(f'accessors/read_from_file/{fname}')

    def test_load_pixel_classifier(self):
        mid = self.assertPutSuccess(
            'ilastik/seg/load/',
            body={'project_file': self._get_models()['pixel_classifier_segmentation']['project_file']},
        )['model_id']
        self.assertTrue(mid.startswith('IlastikPixelClassifierModel'))
        return mid

    def test_load_object_classifier(self):
        mid = self.assertPutSuccess(
            'ilastik/seg_to_obj/load/',
            body={'project_file': self._get_models()['object_classifier']['project_file']},
        )['model_id']
        self.assertTrue(mid.startswith('IlastikObjectClassifierFromSegmentationModel'))
        return mid

    def _object_map_workflow(self, ob_classifer_id):
        oid = self.assertPutSuccess(
            'pipelines/roiset_to_obmap/infer',
            body={
                'accessor_id': self.test_load_input_accessor(),
                'pixel_classifier_segmentation_model_id': self.test_load_pixel_classifier(),
                'object_classifier_model_id': ob_classifer_id,
                'segmentation': {'channel': 0},
                'patches_channel': 1,
                'roi_params': self._get_roi_params(),
                'export_params': self._get_export_params(),
            },
        )['output_accessor_id']
        obmap_fn = self.assertPutSuccess(f'/accessors/write_to_file/{oid}')
        where_out = self.assertGetSuccess('paths')['outbound_images']
        obmap_fp = Path(where_out) / obmap_fn
        self.assertTrue(obmap_fp.exists())
        return generate_file_accessor(obmap_fp)

    def test_workflow_with_object_classifier(self):
        acc = self._object_map_workflow(self.test_load_object_classifier())
        self.assertTrue(np.all(acc.unique()[0] == [0, 1, 2]))

    def test_workflow_without_object_classifier(self):
        acc = self._object_map_workflow(None)
        self.assertTrue(np.all(acc.unique()[0] == [0, 1]))
